#!/usr/bin/env python3
"""
Analyze docstrings to detect errors.

If no argument is provided, it does a quick check of docstrings and returns
a csv with all API functions and results of basic checks.

If a function or method is provided in the form "riptable.function",
"riptable.module.class.method", etc. a list of all errors in the docstring for
the specified function or method.

Usage::
    $ ./validate_docstrings.py
    $ ./validate_docstrings.py riptable.FastArray.get_name
"""

# Adapted from https://github.com/pandas-dev/pandas/blob/main/scripts/validate_docstrings.py
# Module walking inspired by https://github.com/pyvista/numpydoc-validation/blob/main/numpydoc_validation/_validate.py

from __future__ import annotations

import argparse
import doctest
import importlib
import inspect
import io
import json
import os
import pathlib
import re
import subprocess
import sys
import tempfile
import typing

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

import matplotlib
import matplotlib.pyplot as plt
import numpy
from numpydoc.docscrape import get_doc_object
from numpydoc.validate import (
    Validator,
    validate,
)
import pandas
import riptable

# With template backend, matplotlib plots nothing
matplotlib.use("template")


ERROR_MSGS = {
    "GL99": "Error parsing docstring: {doc_parse_error}",
    "GL98": "Private classes ({mentioned_private_classes}) should not be " "mentioned in public docstrings",
    "GL97": "Use 'array-like' rather than 'array_like' in docstrings.",
    "SA99": "{reference_name} in `See Also` section does not need `riptable` " "prefix, use {right_reference} instead.",
    "EX99": "Examples do not pass tests:\n{doctest_log}",
    "EX98": "flake8 error {error_code}: {error_message}{times_happening}",
    "EX97": "Do not import {imported_library}, as it is imported automatically for the examples",
    "EX96": "flake8 warning {error_code}: {error_message}{times_happening}",
}

OUT_FORMAT_OPTS = "default", "json", "actions"

NAMES_FROM_OPTS = "module", "rst"

IGNORE_VALIDATION = {
    # Styler methods are Jinja2 objects who's docstrings we don't own.
    # "Styler.env",
    # "Styler.template_html",
    # "Styler.template_html_style",
    # "Styler.template_html_table",
    # "Styler.template_latex",
    # "Styler.template_string",
    # "Styler.loader",
}

PRIVATE_CLASSES = [
    # "NDFrame",
    # "IndexOpsMixin",
]

IMPORT_CONTEXT = {
    "np": numpy,
    "pd": pandas,
    "rt": riptable,
}


def riptable_error(code, **kwargs):
    """
    Copy of the numpydoc error function, since ERROR_MSGS can't be updated
    with our custom errors yet.
    """
    return (code, ERROR_MSGS[code].format(**kwargs))


def get_api_items(api_doc_fd):
    """
    Yield information about all public API items.

    Parse api.rst file from the documentation, and extract all the functions,
    methods, classes, attributes... This should include all riptable public API.

    Parameters
    ----------
    api_doc_fd : file descriptor
        A file descriptor of the API documentation page, containing the table
        of contents with all the public API.

    Yields
    ------
    name : str
        The name of the object (e.g. 'riptable.FastArray.get_name).
    func : function
        The object itself. In most cases this will be a function or method,
        but it can also be classes, properties, cython objects...
    section : str
        The name of the section in the API page where the object item is
        located.
    subsection : str
        The name of the subsection in the API page where the object item is
        located.
    """
    current_module = "riptable"
    previous_line = current_section = current_subsection = ""
    position = None
    for line in api_doc_fd:
        line = line.strip()
        if len(line) == len(previous_line):
            if set(line) == set("-"):
                current_section = previous_line
                continue
            if set(line) == set("~"):
                current_subsection = previous_line
                continue

        if line.startswith(".. currentmodule::"):
            current_module = line.replace(".. currentmodule::", "").strip()
            continue

        if line == ".. autosummary::":
            position = "autosummary"
            continue

        if position == "autosummary":
            if line == "":
                position = "items"
                continue

        if position == "items":
            if line == "":
                position = None
                continue
            item = line.strip()
            if item in IGNORE_VALIDATION:
                continue
            func = importlib.import_module(current_module)
            for part in item.split("."):
                func = getattr(func, part)

            yield (
                ".".join([current_module, item]),
                func,
                current_section,
                current_subsection,
            )

        previous_line = line


class RiptableDocstring(Validator):
    def __init__(self, func_name: str, doc_obj=None) -> None:
        self.func_name = func_name
        if doc_obj is None:
            doc_obj = get_doc_object(Validator._load_obj(func_name))
        super().__init__(doc_obj)

    @property
    def name(self):
        return self.func_name

    @property
    def mentioned_private_classes(self):
        return [klass for klass in PRIVATE_CLASSES if klass in self.raw_doc]

    @property
    def examples_errors(self):
        flags = doctest.NORMALIZE_WHITESPACE | doctest.IGNORE_EXCEPTION_DETAIL
        finder = doctest.DocTestFinder()
        runner = doctest.DocTestRunner(optionflags=flags)
        error_msgs = ""
        current_dir = set(os.listdir())
        for test in finder.find(self.raw_doc, self.name, globs=IMPORT_CONTEXT):
            f = io.StringIO()
            runner.run(test, out=f.write)
            error_msgs += f.getvalue()
        leftovers = set(os.listdir()).difference(current_dir)
        if leftovers:
            for leftover in leftovers:
                path = pathlib.Path(leftover).resolve()
                if path.is_dir():
                    path.rmdir()
                elif path.is_file():
                    path.unlink(missing_ok=True)
            error_msgs += (
                f"The following files were leftover from the doctest: " f"{leftovers}. Please use # doctest: +SKIP"
            )
        return error_msgs

    @property
    def examples_source_code(self):
        lines = doctest.DocTestParser().get_examples(self.raw_doc)
        return [line.source for line in lines]

    def validate_pep8(self):
        if not self.examples:
            return

        # F401 is needed to not generate flake8 errors in examples
        # that do not use the imported context
        content = ""
        for k, v in IMPORT_CONTEXT.items():
            content += f"import {v.__name__} as {k}  # noqa: F401\n"
        content += "".join((*self.examples_source_code,))

        error_messages = []
        try:
            fd, fname = tempfile.mkstemp(prefix="val-", suffix=".py")
            file = os.fdopen(fd, mode="w", encoding="utf-8")
            file.write(content)
            file.close()
            cmd = ["python", "-m", "flake8", "--quiet", "--statistics", fname]
            response = subprocess.run(cmd, capture_output=True, text=True)
            stdout = response.stdout
            stdout = stdout.replace(fname, "")
            messages = stdout.strip("\n")
            if messages:
                error_messages.append(messages)
        finally:
            os.remove(fname)

        for error_message in error_messages:
            error_count, error_code, message = error_message.split(maxsplit=2)
            yield error_code, message, int(error_count)

    def non_hyphenated_array_like(self):
        return "array_like" in self.raw_doc


def riptable_validate(
    func_name: str,
    errors: typing.Optional(list[str]) = None,
    not_errors: typing.Optional(list[str]) = None,
    flake8_errors: typing.Optional(list[str]) = None,
):
    """
    Call the numpydoc validation, and add the errors specific to riptable.

    Parameters
    ----------
    func_name : str
        Name of the object of the docstring to validate.

    Returns
    -------
    dict
        Information about the docstring and the errors found.
    """
    func_obj = Validator._load_obj(func_name)
    doc_parse_error = None
    try:
        doc_obj = get_doc_object(func_obj, doc=func_obj.__doc__)
    except ValueError as ex:
        doc_parse_error = str(ex)
        doc_obj = get_doc_object(func_obj, doc="")
    doc = RiptableDocstring(func_name, doc_obj)
    result = validate(doc_obj)
    if doc_parse_error:
        result["errors"].insert(0, riptable_error("GL99", doc_parse_error=doc_parse_error))

    mentioned_errs = doc.mentioned_private_classes
    if mentioned_errs:
        result["errors"].append(riptable_error("GL98", mentioned_private_classes=", ".join(mentioned_errs)))

    if doc.see_also:
        for rel_name in doc.see_also:
            if rel_name.startswith("riptable."):
                result["errors"].append(
                    riptable_error(
                        "SA99",
                        reference_name=rel_name,
                        right_reference=rel_name[len("riptable.") :],
                    )
                )

    def matches(test: str, matches: list[str]):
        for match in matches:
            if test.startswith(match):
                return True
        return False

    result["examples_errs"] = ""
    if doc.examples:
        result["examples_errs"] = doc.examples_errors
        if result["examples_errs"]:
            result["errors"].append(riptable_error("EX99", doctest_log=result["examples_errs"]))

        for error_code, error_message, error_count in doc.validate_pep8():
            times_happening = f" ({error_count} times)" if error_count > 1 else ""
            result["errors"].append(
                riptable_error(
                    "EX98" if flake8_errors and matches(error_code, flake8_errors) else "EX96",
                    error_code=error_code,
                    error_message=error_message,
                    times_happening=times_happening,
                )
            )
        examples_source_code = "".join(doc.examples_source_code)
        for wrong_import in [v.__name__ for v in IMPORT_CONTEXT.values()]:
            if re.search(f"import {wrong_import}\W+", examples_source_code):
                result["errors"].append(riptable_error("EX97", imported_library=wrong_import))

    if doc.non_hyphenated_array_like():
        result["errors"].append(riptable_error("GL97"))

    plt.close("all")

    if errors or not_errors:
        filtered_errors = []
        for err_code, err_desc in result["errors"]:
            if not (errors and not matches(err_code, errors) or not_errors and matches(err_code, not_errors)):
                filtered_errors.append((err_code, err_desc))
        result["errors"] = filtered_errors

    return result


def get_all_objects(root: object) -> set[object]:
    objs = set()
    for name, obj in inspect.getmembers(root):
        # ignore any private names
        if name.startswith("_"):
            continue
        # ignore if obj is not class or routine
        if not inspect.isclass(obj) and not inspect.isroutine(obj):
            continue
        objs.add(obj)
        if inspect.isclass(obj):
            objs |= get_all_objects(obj)
    return objs


def get_module_items(modulename) -> list[str]:
    module = importlib.import_module(modulename)
    items = []
    for obj in get_all_objects(module):
        # extract the obj full name, ignoring anything not named
        try:
            fullname = obj.__module__ + "." + obj.__qualname__
        except (AttributeError, TypeError):
            continue
        # ignore any objects not part of this module
        if not modulename in fullname:
            continue
        items.append((fullname, obj, None, None))
    return items


def validate_all(
    match: str,
    not_match: str = None,
    names_from: str = NAMES_FROM_OPTS[0],
    errors: typing.Optional(list[str]) = None,
    not_errors: typing.Optional(list[str]) = None,
    flake8_errors: typing.Optional(list[str]) = None,
    ignore_deprecated: bool = False,
    verbose: int = 0,
) -> dict:
    """
    Execute the validation of all docstrings, and return a dict with the
    results.

    Parameters
    ----------
    prefix : str or None
        If provided, only the docstrings that start with this pattern will be
        validated. If None, all docstrings will be validated.
    ignore_deprecated: bool, default False
        If True, deprecated objects are ignored when validating docstrings.

    Returns
    -------
    dict
        A dictionary with an item for every function/method... containing
        all the validation information.
    """
    result = {}
    seen = {}

    api_items = []

    if names_from == "rst":
        base_path = pathlib.Path(__file__).parent.parent
        api_doc_fnames = pathlib.Path(base_path, "doc", "source", "reference")
        for api_doc_fname in api_doc_fnames.glob("*.rst"):
            with open(api_doc_fname) as f:
                api_items += list(get_api_items(f))
    else:
        api_items.extend(get_module_items("riptable"))

    api_items.sort(key=lambda v: v[0])

    match_re = re.compile(match) if match else None
    not_match_re = re.compile(not_match) if not_match else None

    for func_name, _, section, subsection in api_items:
        if match_re and not match_re.search(func_name) or not_match_re and not_match_re.search(func_name):
            if verbose > 1:
                print(f"Ignoring {func_name} not matching prefix {match}")
            continue
        if verbose:
            print(f"Validating {func_name}... ", end="", flush=True)
        doc_info = riptable_validate(
            func_name,
            errors=errors,
            not_errors=not_errors,
            flake8_errors=flake8_errors,
        )
        if ignore_deprecated and doc_info["deprecated"]:
            if verbose > 1:
                print(f"Ignoring deprecated {func_name}")
            continue
        if verbose:
            status = "FAILED" if len(doc_info["errors"]) else "OK"
            print(status)
        result[func_name] = doc_info

        shared_code_key = doc_info["file"], doc_info["file_line"]
        shared_code = seen.get(shared_code_key, "")
        result[func_name].update(
            {
                "in_api": True,
                "section": section,
                "subsection": subsection,
                "shared_code_with": shared_code,
            }
        )

        seen[shared_code_key] = func_name

    return result


def print_validate_all_results(
    match: str,
    not_match: str = None,
    names_from: str = NAMES_FROM_OPTS[0],
    errors: typing.Optional(list[str]) = None,
    not_errors: typing.Optional(list[str]) = None,
    flake8_errors: typing.Optional(list[str]) = None,
    out_format: str = OUT_FORMAT_OPTS[0],
    ignore_deprecated: bool = False,
    outfile: typing.IO = sys.stdout,
    verbose: int = 0,
):
    if out_format not in OUT_FORMAT_OPTS:
        raise ValueError(f'Unknown output_format "{out_format}"')

    result = validate_all(
        match,
        not_match=not_match,
        names_from=names_from,
        errors=errors,
        not_errors=not_errors,
        flake8_errors=flake8_errors,
        ignore_deprecated=ignore_deprecated,
        verbose=verbose,
    )

    if verbose:
        print("Results:")

    if out_format == "json":
        json.dump(result, outfile, indent=2)
    else:
        prefix = "##[error]" if out_format == "actions" else ""
        for name, res in result.items():
            for err_code, err_desc in res["errors"]:
                outfile.write(f'{prefix}{res["file"]}:{res["file_line"]}: {name}: {err_code}: {err_desc}\n')

    exit_status = 0
    for name, res in result.items():
        if len(res["errors"]):
            exit_status = 1
            break

    if verbose:
        print("Validation " + ("OK!" if exit_status == 0 else "FAILED!"))

    return exit_status


def print_validate_one_results(
    func_name: str,
    errors: typing.Optional(list[str]) = None,
    not_errors: typing.Optional(list[str]) = None,
    flake8_errors: typing.Optional(list[str]) = None,
    outfile: typing.IO = sys.stdout,
    verbose: int = 0,
):
    def header(title, width=80, char="#"):
        full_line = char * width
        side_len = (width - len(title) - 2) // 2
        adj = "" if len(title) % 2 == 0 else " "
        title_line = f"{char * side_len} {title}{adj} {char * side_len}"

        return f"\n{full_line}\n{title_line}\n{full_line}\n\n"

    if verbose:
        print(f"Validating {func_name}...")
    result = riptable_validate(
        func_name,
        errors=errors,
        not_errors=not_errors,
        flake8_errors=flake8_errors,
    )

    outfile.write(header(f"Docstring ({func_name})"))
    outfile.write(f"{result['docstring']}\n")

    outfile.write(header("Validation"))
    if result["errors"]:
        outfile.write(f'{len(result["errors"])} Errors found:\n')
        for err_code, err_desc in result["errors"]:
            if err_code == "EX99":  # Failing examples are printed at the end
                outfile.write("\tExamples do not pass tests\n")
                continue
            outfile.write(f"\t{err_code}: {err_desc}\n")
    else:
        outfile.write(f'Docstring for "{func_name}" correct. :)\n')

    if result["examples_errs"]:
        outfile.write(header("Doctests"))
        outfile.write(result["examples_errs"])


def find_parent_dir_containing(filename: str) -> typing.Optional[str]:
    cur_dir = os.getcwd()
    while not os.path.exists(os.path.join(cur_dir, filename)):
        cur_dir = os.path.dirname(cur_dir)
        if not cur_dir:
            return None
    return cur_dir


def find_pyproject_toml() -> typing.Optional[str]:
    pyproj_toml_filename = "pyproject.toml"
    root_dir = find_parent_dir_containing(pyproj_toml_filename)
    return os.path.join(root_dir, pyproj_toml_filename) if root_dir else None


def main():
    """
    Main entry point. Call the validation for one or for all docstrings.
    """
    argparser = argparse.ArgumentParser(description="Validate riptable docstrings")
    argparser.add_argument(
        "function",
        nargs="?",
        default=None,
        help="Function or method to validate (e.g. riptable.FastArray.get_name) "
        "if not provided, all docstrings are validated and returned "
        "as JSON.",
    )
    argparser.add_argument(
        "--names-from",
        default=NAMES_FROM_OPTS[0],
        choices=NAMES_FROM_OPTS,
        help=f"Source of names when searching for all docstrings. It can be one of {str(NAMES_FROM_OPTS)[1:-1]} (default: '%(default)s').",
    )
    argparser.add_argument(
        "--format",
        default=OUT_FORMAT_OPTS[0],
        choices=OUT_FORMAT_OPTS,
        help="format of the output when validating "
        "multiple docstrings (ignored when validating one). "
        f"It can be one of {str(OUT_FORMAT_OPTS)[1:-1]} (default: '%(default)s').",
    )
    argparser.add_argument(
        "--match",
        default=None,
        help="Regex pattern for matching "
        "docstring names, in order to decide which ones "
        'will be validated. The match "FastArray" '
        "will make the script validate all the docstrings "
        'of methods containing "FastArray". It is '
        "ignored if function option is provided.",
    )
    argparser.add_argument(
        "--not-match",
        default=None,
        help="Regex pattern for not matching "
        "docstring names, in order to decide which ones "
        'will be validated. The not-match "mapping" '
        "will make the script validate all the docstrings "
        'of methods not containing "mapping". '
        "Any matches are performed first, then any not-matches are excluded. "
        "It is ignored if function option is provided.",
    )
    argparser.add_argument(
        "--errors",
        default=None,
        help="Comma separated "
        "list of error codes to validate. By default it "
        "validates all errors. Ignored when validating "
        "a single docstring.",
    )
    argparser.add_argument(
        "--not-errors",
        default=None,
        help="Comma separated "
        "list of error codes not to validate. Empty by default. "
        "Ignored when validating a single docstring.",
    )
    argparser.add_argument(
        "--ignore-deprecated",
        default=False,
        action="store_true",
        help="If this flag is set, deprecated objects are ignored when validating all docstrings.",
    )
    argparser.add_argument(
        "--flake8-errors",
        default=None,
        help="Comma separated list of flake8 error codes to treat as EX98 errors. Others are treated as warnings.",
    )
    argparser.add_argument(
        "--out",
        "-o",
        default=None,
        type=str,
        help="Output file path, else use stdout.",
    )
    argparser.add_argument(
        "--verbose",
        "-v",
        default=0,
        action="count",
        help="Emit verbose progress output. Specify multiple times for more verbosity.",
    )

    args = argparse.Namespace()

    pyproj_toml_path = find_pyproject_toml()
    if pyproj_toml_path:
        with open(pyproj_toml_path, "rb") as f:
            pyproj_toml = tomllib.load(f)
        config = pyproj_toml.get("tool", {}).get("validate_docstrings", {})
        for k, v in config.items():
            setattr(args, k, v)

    argparser.parse_args(namespace=args)

    with open(args.out, "w", encoding="utf-8", errors="backslashreplace") if args.out else open(
        sys.stdout.fileno(), "w", closefd=False
    ) as outfile:
        errors = args.errors.split(",") if args.errors is not None else None
        not_errors = args.not_errors.split(",") if args.not_errors is not None else None
        flake8_errors = args.flake8_errors.split(",") if args.flake8_errors is not None else None

        if args.function is None:
            return print_validate_all_results(
                match=args.match,
                not_match=args.not_match,
                names_from=args.names_from,
                errors=errors,
                not_errors=not_errors,
                flake8_errors=flake8_errors,
                out_format=args.format,
                ignore_deprecated=args.ignore_deprecated,
                outfile=outfile,
                verbose=args.verbose,
            )
        else:
            print_validate_one_results(
                args.function,
                errors=errors,
                not_errors=not_errors,
                flake8_errors=flake8_errors,
                outfile=outfile,
                verbose=args.verbose,
            )
            return 0


if __name__ == "__main__":
    sys.exit(main())



import os

import pytest

from riptable import Dataset, Multiset, Struct

from .vnu_checker import VNUChecker


@pytest.mark.xfail(reason="Requires the VNU.jar HTML validator")
@pytest.mark.parametrize(
    "rt_obj",
    [
        # Add complex examples that use composition and recursion of other riptable data structures.
        Dataset({"one": 1}),
        Multiset({"ds": Dataset({"one": 1})}),
        Struct({"ds": Dataset({"one": 1})}),
    ],
)
def test_simple_html_repr(rt_obj, tmp_path):
    # dump HTML representation to files
    if not os.path.isdir(tmp_path):
        tmp_path.mkdir()
    checker = VNUChecker(base_path=str(tmp_path), errors_only=True, ascii_quotes=True)
    fn = os.path.join(checker.base_path, rt_obj.__class__.__name__ + ".html")
    with open(fn, "w") as f:
        # Consider using builtin library lxml or third-party BeautifulSoup for better HTML dumps:
        # lxml:
        # doc = html.fromstring(rt_obj._repr_html_())
        # f.write(etree.tostring(doc, encoding="unicode", pretty_print=True))
        # BeautifulSoup:
        # f.write(BeautifulSoup(rt_obj._repr_html_(), features="lxml").prettify())
        f.write(rt_obj._repr_html_())
    if hasattr(rt_obj, "_T"):  # dump and validate transpose representation
        with open(fn, "w") as f:
            f.write(rt_obj._repr_html_())

    # validate the dumped HTML files are valid HTML
    errors = checker.validate()

    assert not len(errors), f"Checker failed.\nChecker command: {checker}\nGot the following errors:\n" + "\n".join(
        errors
    )



import unittest
from contextlib import contextmanager
from enum import IntEnum
from typing import List, Union

import pytest
from IPython import get_ipython
from IPython.core.completer import provisionalcompleter
from IPython.terminal import ptutils

from riptable import Categorical, Dataset, FastArray, Multiset, Struct, arange
from riptable.rt_misc import autocomplete
from riptable.tests.test_categorical import decision_dict
from riptable.tests.utils import LikertDecision
from riptable.Utils.ipython_utils import (
    _get_key_names,
    enable_custom_attribute_completion,
)

_DEBUG = False


class CompletionType(IntEnum):
    UNDEFINED = 0
    KEY = 1
    ATTRIBUTE = 2


class GreedyJediConfigType(IntEnum):
    NEITHER = 0
    GREEDY = 1
    JEDI = 2
    GREEDY_JEDI = 3


_COMPLETION_TYPE_TABLE = [CompletionType.KEY, CompletionType.ATTRIBUTE]


_GREEDY_JEDI_CONFIG_TABLE = [
    GreedyJediConfigType.NEITHER,
    GreedyJediConfigType.GREEDY,
    GreedyJediConfigType.JEDI,
    GreedyJediConfigType.GREEDY_JEDI,
]


CODES = [1, 44, 44, 133, 75]


# Todo - put this in a function generator that can get data based on data type
# Add data entries to _RT_DATA_TABLE in an append only way since tests depend on ordering.
# Tests should not depend on ordering of _RT_DATA_TABLE, see above work item.
_RT_DATA_TABLE = [
    Categorical(
        FastArray(["a", "b", "c", "c", "d", "a", "b"]),
        ordered=True,
        base_index=1,
        filter=None,
    ),
    Categorical(CODES, LikertDecision),
    Categorical(CODES, decision_dict),
    Categorical(["b", "a", "a", "c", "a", "b"], ["b", "a", "c", "e"], sort_gb=True),
    Dataset(
        {
            _k: list(range(_i * 10, (_i + 1) * 10))
            for _i, _k in enumerate(
                [
                    "alpha",
                    "beta",
                    "gamma",
                    "delta",
                    "epsilon",
                    "zeta",
                    "eta",
                    "theta",
                    "iota",
                    "kappa",
                    "lambada",
                    "mu",
                    "nu",
                    "xi",
                    "omnicron",
                    "pi",
                ]
            )
        }
    ),
    Multiset(
        {
            "ds_alpha": Dataset(
                {
                    k: list(range(i * 10, (i + 1) * 10))
                    for i, k in enumerate(["alpha", "beta", "gamma", "delta", "epsilon", "zeta"])
                }
            ),
            "ds_beta": Dataset(
                {
                    k: list(range(i * 10, (i + 1) * 10))
                    for i, k in enumerate(["eta", "theta", "iota", "kappa", "lambada", "mu"])
                }
            ),
        }
    ),
    Struct(
        {
            "alpha": 1,
            "beta": [2, 3],
            "gamma": ["2", "3"],
            "delta": arange(10),
            "epsilon": Struct(
                {
                    "theta": Struct(
                        {
                            "kappa": 3,
                            "zeta": 4,
                        }
                    ),
                    "iota": 2,
                }
            ),
        }
    ),
]


@contextmanager
def greedy_completion():
    ip = get_ipython()
    greedy_original = ip.Completer.greedy
    try:
        ip.Completer.greedy = True
        yield
    finally:
        ip.Completer.greedy = greedy_original


@contextmanager
def jedi_completion():
    ip = get_ipython()
    use_jedi_original = ip.Completer.use_jedi
    try:
        ip.Completer.use_jedi = True
        yield
    finally:
        ip.Completer.use_jedi = use_jedi_original


def get_completion_text(completion_type) -> str:
    text: str = ""
    if completion_type == CompletionType.KEY:
        text = "d['"
    elif completion_type == CompletionType.ATTRIBUTE:
        text = "d."
    else:
        raise ValueError(
            "get_completion_text: could not handled {} of type {}".format(completion_type, type(completion_type))
        )
    return text


@pytest.mark.parametrize("greedy_jedi_config_type", _GREEDY_JEDI_CONFIG_TABLE)
@pytest.mark.parametrize("completion_type", _COMPLETION_TYPE_TABLE)
@pytest.mark.parametrize("rt_data", _RT_DATA_TABLE)
def test_simple_completion(
    greedy_jedi_config_type: GreedyJediConfigType,
    completion_type: CompletionType,
    rt_data: Union[FastArray, Dataset, Categorical, Struct],
):
    ip = get_ipython()
    ip.user_ns["d"] = rt_data

    completions: list
    text = get_completion_text(completion_type)
    with provisionalcompleter():
        if greedy_jedi_config_type == GreedyJediConfigType.GREEDY_JEDI:
            with greedy_completion():
                with jedi_completion():
                    completions = list(ip.Completer.completions(text, len(text)))
        elif greedy_jedi_config_type == GreedyJediConfigType.GREEDY:
            with greedy_completion():
                completions = list(ip.Completer.completions(text, len(text)))
        elif greedy_jedi_config_type == GreedyJediConfigType.JEDI:
            with jedi_completion():
                completions = list(ip.Completer.completions(text, len(text)))
        elif greedy_jedi_config_type == GreedyJediConfigType.NEITHER:
            completions = list(ip.Completer.completions(text, len(text)))
        else:
            raise ValueError(
                "test_simple_completion: could not handle greedy_jedi_config_type {} of type {}".format(
                    greedy_jedi_config_type, type(greedy_jedi_config_type)
                )
            )

    matches = [completion.text for completion in completions]
    if isinstance(rt_data, Categorical):
        rt_data = rt_data.categories()
    if _get_key_names(rt_data):
        for key in rt_data.keys():
            assert True, key in matches


# add struct to categorical, dataset, fastarray, and struct completion tests


class TestIPCompleterIntegration(unittest.TestCase):
    def setUp(self):
        """
        Silence all PendingDeprecationWarning when testing the completer integration.
        PendingDeprecationWarning: `Completer.complete` is pending deprecation since IPython 6.0 and will be replaced by `Completer.completions`.
        """
        self._assertwarns = self.assertWarns(PendingDeprecationWarning)
        self._assertwarns.__enter__()

    def tearDown(self):
        try:
            self._assertwarns.__exit__(None, None, None)
        except AssertionError:
            pass

    # region Categorical
    def test_categorical_string_array_key_completion(self):
        ip = get_ipython()
        complete = ip.Completer.complete
        lst = ["a", "b", "c", "c", "d", "a", "b"]  # type: List[str]
        ip.user_ns["cat"] = Categorical(FastArray(lst), ordered=True, base_index=1, filter=None)
        _, matches = complete(line_buffer="cat['")
        for s in lst:
            self.assertIn(s, matches)

    def test_categorical_int_enum_key_completion(self):
        ip = get_ipython()
        complete = ip.Completer.complete

        ip.user_ns["cat"] = Categorical(CODES, LikertDecision)
        _, matches = complete(line_buffer="cat['")
        for k in decision_dict.keys():
            self.assertIn(k, matches)

    def test_categorical_dict_key_completion(self):
        ip = get_ipython()
        complete = ip.Completer.complete

        ip.user_ns["cat"] = Categorical(CODES, decision_dict)
        _, matches = complete(line_buffer="cat['")
        for k in decision_dict.keys():
            self.assertIn(k, matches)

    def test_categorical_numeric_array_key_completion(self):
        ip = get_ipython()
        complete = ip.Completer.complete

        lst = [1, 44, 44, 133, 75]  # type: List[int]
        ip.user_ns["cat"] = Categorical(FastArray(lst))
        _, matches = complete(line_buffer="cat['")
        expected = [str(i) for i in lst]
        for c in expected:
            self.assertIn(c, matches)

    # todo alz 20191202 - add variant test cases for types of various combinations of input keys
    def test_categorical_multi_key_completion(self):
        ip = get_ipython()
        complete = ip.Completer.complete

        # note - 'e' is not in first list
        lst1 = ["b", "a", "a", "c", "a", "b"]  # type: List[str]
        lst2 = ["b", "a", "c", "e"]  # type: List[str]
        ip.user_ns["cat"] = Categorical(lst1, lst2, sort_gb=True)
        _, matches = complete(line_buffer="cat['")
        for c in lst1:
            self.assertIn(c, matches)
        for c in lst2:
            self.assertIn(c, matches)

    # todo alz 20191202 - add test cases for column selection cat[[string1, string2]]
    # endregion


# region rt_misc autocomplete
@pytest.mark.parametrize(
    "rt_data, text",
    [
        # Simple attribute completion tests
        (_RT_DATA_TABLE[4], "d."),
        (_RT_DATA_TABLE[5], "d."),
        (_RT_DATA_TABLE[6], "d."),
        # Nested attribute completion tests
        # Multiset
        (_RT_DATA_TABLE[5], "m."),
        (_RT_DATA_TABLE[5], "m.ds_alpha."),
        # Struct
        (_RT_DATA_TABLE[6], "s."),
        (_RT_DATA_TABLE[6], "s.epsilon."),
        (_RT_DATA_TABLE[6], "s.epsilon.theta."),
    ],
)
def test_riptable_autocomplete_keys_ordering(rt_data: Union[Dataset, Multiset, Struct], text: str):
    # Keys are expected to be in sorted order.
    autocomplete()
    with provisionalcompleter(), jedi_completion():
        ip = get_ipython()

        text_split = text.split(".")
        ns_key: str = text_split[0]
        ip.user_ns[ns_key] = rt_data

        completions = list(ptutils._deduplicate_completions(text, ip.Completer.completions(text, len(text))))
        matches = [completion.text for completion in completions]

        expected_matches: List[str] = []
        if len(text_split) > 2:  # support arbitrary object depth expectations
            locals()[ns_key] = rt_data
            rt_data = eval(text[:-1])

        if _get_key_names(rt_data):  # support object key expectations
            expected_matches.extend(sorted(rt_data.keys()))

        assert len(expected_matches), f"type {type(rt_data)} has no keys"

        for i, k in enumerate(expected_matches):
            # Match will be key, attribute, or method name.
            # E.g., given namespace name `d` and key `alpha` completion matches will contain the value `alpha`.
            assert i == matches.index(
                k
            ), f"expected key {k} in completion result position {i}\nwanted keys {expected_matches}\ngot matches {matches}"


@pytest.mark.parametrize(
    "rt_data, text",
    [
        # Simple attribute completion tests
        (_RT_DATA_TABLE[4], "d."),
        (_RT_DATA_TABLE[5], "d."),
        (_RT_DATA_TABLE[6], "d."),
        # Nested attribute completion tests
        # Multiset
        (_RT_DATA_TABLE[5], "m."),
        (_RT_DATA_TABLE[5], "m.ds_alpha."),
        # Struct
        (_RT_DATA_TABLE[6], "s."),
        (_RT_DATA_TABLE[6], "s.epsilon."),
        (_RT_DATA_TABLE[6], "s.epsilon.theta."),
    ],
)
def test_riptable_autocomplete_equivalent_to_ipython_without_ordering(
    rt_data: Union[Dataset, Multiset, Struct], text: str
):
    # Keys are expected to be in sorted order, but are not. This is tracked in RIP-323.
    # Ensure riptable autocomplete contains sane completion results as IPython's auto completion.
    with provisionalcompleter(), jedi_completion():
        ip = get_ipython()

        ns_key: str = text.split(".")[0]
        ip.user_ns[ns_key] = rt_data

        expected_completions = list(ptutils._deduplicate_completions(text, ip.Completer.completions(text, len(text))))
        expected_matches = set([completion.text for completion in expected_completions])

        autocomplete()

        completions = list(ptutils._deduplicate_completions(text, ip.Completer.completions(text, len(text))))
        matches = set([completion.text for completion in completions])

        # IPython's set of completions should be the same as Riptable's set of completions.
        assert not expected_matches.symmetric_difference(
            matches
        ), f"expected matches {expected_matches}\ngot matches {matches}"


# endregion rt_misc autocomplete


# region ipython_utils enable_custom_attribute_completion
# Please leave these tests at the end since they activate the custom completion machinery.
# As of 20191218, if ``CUSTOM_COMPLETION`` is toggled on it results in a one-time registration of custom
# but does not support deregistration.
@pytest.mark.parametrize(
    "rt_data, text",
    [
        # Simple attribute completion tests
        (_RT_DATA_TABLE[4], "d."),
        (_RT_DATA_TABLE[5], "d."),
        (_RT_DATA_TABLE[6], "d."),
        # Nested attribute completion tests
        # Multiset
        (_RT_DATA_TABLE[5], "m."),
        (_RT_DATA_TABLE[5], "m.ds_alpha."),
        # Struct
        (_RT_DATA_TABLE[6], "s."),
        (_RT_DATA_TABLE[6], "s.epsilon."),
        (_RT_DATA_TABLE[6], "s.epsilon.theta."),
    ],
)
def test_monkey_patch_complete_match_ordering(rt_data: Union[Dataset, Multiset, Struct], text: str):
    # 1) Keys are expected to be in sorted order.
    enable_custom_attribute_completion()

    ip = get_ipython()
    complete = ip.Completer.complete

    text_split = text.split(".")
    ns_key: str = text_split[0]
    ip.user_ns[ns_key] = rt_data

    _, matches = complete(line_buffer=text)

    expected_matches: List[str] = []
    if len(text_split) > 2:  # support arbitrary object depth expectations
        locals()[ns_key] = rt_data
        rt_data = eval(text[:-1])

    if _get_key_names(rt_data):  # support object key expectations
        expected_matches.extend(sorted(rt_data.keys()))

    for expected_index, expected_match in enumerate(expected_matches):
        # Match will be namespace name dot key, attribute, or method name.
        # E.g., given namespace name `d` and key `alpha` completion matches will contain the value `d.alpha`.
        expected_match: str = text + expected_match
        assert expected_index == matches.index(
            expected_match
        ), f"expected match {expected_match} in completion result position {expected_index}\nwanted keys {expected_matches}\ngot matches {matches}"


# endregion ipython_utils enable_custom_attribute_completion


if __name__ == "__main__":
    tester = unittest.main()



# -*- coding: utf-8 -*-
from typing import Callable

import numpy as np
import riptide_cpp as rc

from ..rt_dataset import Dataset
from ..rt_fastarray import FastArray
from ..rt_fastarray import FastArray as FA
from ..rt_multiset import Multiset
from ..rt_numpy import empty
from .analysis import _analyze_results, enable_bench_estimators

# TODO:  When running in Py37+, add time.perf_counter_ns() (and maybe time.time_ns()).
_timestamp_funcs = {"get_nano_time": rc.GetNanoTime, "tsc": rc.GetTSC}
"""
Table of functions for getting a nanosecond-precision timestamp.
"""


timestamper = _timestamp_funcs["get_nano_time"]
"""The timestamping function to use in benchmarks."""


def _makename(dtype, asize):
    nm = dtype.__name__
    last = len(nm)
    while last >= 0:
        last = last - 1
        chr = nm[last]
        if chr < "0" or chr > "9":
            break
    if last > 0 and last < len(nm) - 1:
        nm = nm[0] + nm[last + 1 :]

    number = str(asize)
    if asize >= 1000:
        number = str(asize // 1000) + "K"
        if asize >= 1_000_000:
            number = str(asize // 1_000_000) + "M"
    return nm + "_" + number


def _time_run(timerfunc: Callable[[], int], func, *args, loops: int, **kwargs) -> np.ndarray:
    """
    Invoke a function to be benchmarked repeatedly with the specified arguments to capture timing information.

    Parameters
    ----------
    timerfunc : callable
        Function returning a timestamp (in nanoseconds) when invoked.
    func : callable
        The function to be benchmarked.
    args
        Arguments to be passed to `func`.
    loops : int
        The number of times to invoke `func`.
    kwargs
        Keyword arguments to be passed to `func`.

    Returns
    -------
    RunResult
        A `RunResult` holding aggregated data about this benchmark run.
    """
    # preallocate array of int64 for nanoseconds
    atimers = np.zeros(loops, dtype=np.int64)

    for i in range(loops):
        start_time = timerfunc()
        func(*args, **kwargs)
        end_time = timerfunc()
        atimers[i] = end_time - start_time

    return atimers


def benchmark(
    funcs,
    dtypes,
    asizes,
    threads,
    loops=10,
    ratio=False,
    recycle=True,
    funcargs=1,
    scalar=False,
):
    """
    riptable benchmarking comparison performance FA vs numpy

    Will return a Dataset or Multiset depending on ratio=True/False.
    Columns will be arraysizes or dtypes across top
    Rows will be func or threads

    Parameters
    ----------
    funcs:
    dtypes:
    asizes:
    threads: list of threads to use
    loops: int, default 10
    ratio: bool, default True
        True to return a Dataset with the ratio of FastArray/numpy (higher is better)
        False to return a MultiSet with the times in seconds.
    recycle:
    funcargs:
    scalar:

    Examples
    --------
    benchmark(dtypes=np.int32, ratio=True)
    benchmark(funcs=np.add, dtypes=np.float32, funcargs=2, ratio=True)

    Notes
    -----
    Better way to test add on small array
    a=np.arange(1000, dtype=np.int32)
    %timeit np.add(a,a)
    %timeit rc.BasicMathTwoInputs((a, a), 1, 7)
    """

    def fix_scalar(s):
        # make sure always a list
        if not isinstance(s, list):
            return [s]
        return s

    funcs = fix_scalar(funcs)
    threads = fix_scalar(threads)
    dtypes = fix_scalar(dtypes)
    asizes = fix_scalar(asizes)

    func_len = len(funcs)
    thread_len = len(threads)
    dtype_len = len(dtypes)
    asize_len = len(asizes)

    rowlen = func_len * thread_len
    ds = {}

    # now make rownames
    threadnames = []
    funcnames = []
    for func in funcs:
        for thread in threads:
            threadnames.append(thread)
            funcnames.append(func.__name__)

    ds["Thread"] = threadnames
    ds["Func"] = funcnames

    # now make the column names
    colnames = []
    for dtype in dtypes:
        for asize in asizes:
            name = _makename(dtype, asize)
            ds[name] = empty(rowlen, dtype=np.float64)

    ds_fa = Dataset(ds)
    ds_fa.label_set_names(["Thread", "Func"])
    ds_np = ds_fa.copy()

    if not recycle:
        FA._ROFF()
        FA._GCNOW()

    # preserve threading modes
    savemode = rc.ThreadingMode(1)
    savethreads = rc.SetThreadWakeUp(8)

    # Get the timestamping function to use.
    # TODO: Make this user-configurable.
    timestamp_func = _timestamp_funcs["get_nano_time"]

    # make two passes
    dscount = 0
    for ds in [ds_fa, ds_np]:
        col = 2
        for dtype in dtypes:
            for asize in asizes:
                row = 0
                # can make random arrays also
                array = np.arange(asize, dtype=dtype)
                if dscount == 0:
                    # convert to FastArray for first run
                    array = array.view(FastArray)

                for func in funcs:
                    for thread in threads:
                        if thread == 1:
                            rc.ThreadingMode(1)
                        else:
                            rc.ThreadingMode(0)
                            # subtract since this is worker threads and we are counting main thread
                            rc.SetThreadWakeUp(thread - 1)
                        if funcargs == 1:
                            raw_timings = _time_run(timestamp_func, func, array, loops=loops)
                        else:
                            if scalar:
                                # use a scalar as second input
                                raw_timings = _time_run(timestamp_func, func, array, array[0], loops=loops)
                            else:
                                raw_timings = _time_run(timestamp_func, func, array, array, loops=loops)

                        # Analyze the raw result data.
                        analysis = _analyze_results(raw_timings)

                        # If robust stats are enabled, print the whole result so we can see what it looks like.
                        if enable_bench_estimators:
                            print(repr(analysis))

                        ds[col][row] = analysis.median_ns
                        row = row + 1
                        # end of thread loop
                    # end of func loop
                col += 1
                # end of asize loop
            # end of dtype loop
        dscount += 1
        # end of ds loop

    rc.ThreadingMode(savemode)
    rc.SetThreadWakeUp(savethreads)

    if not recycle:
        FA._RON()

    if ratio:
        # return a ratio
        return ds_np / ds_fa

    ms = Multiset({"FastA": ds_fa, "Numpy": ds_np})
    ms.label_set_names(["Thread", "Func"])
    return ms



import math
import re
from typing import Tuple

import pytest

parametrize = pytest.mark.parametrize

from numpy.testing import assert_array_equal

import riptable as rt
from riptable import *

from ..testing.array_assert import assert_array_or_cat_equal

SYMBOLS = ["AAPL", "AMZN", "FB", "GOOG", "IBM"]
PARALLEL_MULTIPLIER = 2000
NB_PARALLEL_SYMBOLS = SYMBOLS * PARALLEL_MULTIPLIER
assert len(NB_PARALLEL_SYMBOLS) >= FAString._APPLY_PARALLEL_THRESHOLD

_top_25_US_cities_by_population_2019_state_names = [
    "New York",
    "California",
    "Illinois",
    "Texas",
    "Arizona",
    "Pennsylvania",
    "Texas",
    "California",
    "Texas",
    "California",
    "Texas",
    "Florida",
    "Texas",
    "Ohio",
    "North Carolina",
    "California",
    "Indiana",
    "Washington",
    "Colorado",
    "District of Columbia",
    "Massachusetts",
    "Texas",
    "Tennessee",
    "Michigan",
    "Oklahoma",
]
"""
The state names (not abbreviations) of the top 25 US cities by population
in 2019 according to the US Census Bureau. Used to provide test data for
unit tests of string functions.
https://en.wikipedia.org/wiki/List_of_United_States_cities_by_population
"""


def _tile_array(arr: np.ndarray, tile_count: int) -> np.ndarray:
    return arr.tile(tile_count) if hasattr(arr, "tile") else np.tile(arr, tile_count)


def _make_parallelizable_array(arr: np.ndarray) -> Tuple[np.ndarray, int]:
    # assumes C-contiguous layout here
    rowcount = arr.shape[0]
    tile_count = int(math.ceil((FAString._APPLY_PARALLEL_THRESHOLD + 1) / rowcount))
    tiled_arr = _tile_array(arr, tile_count)
    return tiled_arr, tile_count


class TestFAString:
    """
    Unit tests for the FAString class (via the ``FastArray.str`` accessor).
    """

    # NOTE: When implementing new unit tests (e.g. for new additions to FAString), make sure to cover
    #       all *combinations* of the following * each applicable input to the function:
    #           * array type: np.ndarray, FastArray, Categorical
    #           * encoding: ascii (bytes), Unicode (str)
    #           * parallelism: serial, parallel (based on the array length and FAString._APPLY_PARALLEL_THRESHOLD).
    #           * empty vs. non-empty array
    #               * for Categorical, this could mean either or both of the following:
    #                   * the Categorical has length zero (e.g. len(my_cat) == 0).
    #                   * the Categorical has no categories, possibly because they were all filtered out.

    cat_symbol = Cat(np.tile(np.arange(len(SYMBOLS) + 1), 3), SYMBOLS)

    def test_cat_base_index_0(self):
        cat = rt.Categorical(np.tile([0, 1], 100), ["abc ", "bcd"], base_index=0)
        result = cat.str.removetrailing()
        expected = rt.Categorical(np.tile([0, 1], 100), np.asarray(["abc", "bcd"]).astype("S4"), base_index=0)
        assert_array_or_cat_equal(result, expected)

    @parametrize("position", [0, 1, -1, -2, 3])
    def test_char(self, position):
        result = FAString(SYMBOLS).char(position)
        expected = rt.FastArray([s[position] if position < len(s) else "" for s in SYMBOLS])
        assert_array_equal(result, expected)

    @parametrize("position", [0, 1, -1, -2])
    def test_char_cat(self, position):
        result = self.cat_symbol.str.char(position)
        expected = Categorical(self.cat_symbol.ikey, [s[position] for s in SYMBOLS], base_index=1)
        assert_array_or_cat_equal(expected, result, relaxed_cat_check=True)
        # check categories are unique
        assert len(set(result.category_array)) == len(result.category_array)

    def test_char_array_position(self):
        position = [-1, 2, 0, 1, 2]
        result = FAString(SYMBOLS).char(position)
        expected = FastArray([s[pos] for s, pos in zip(SYMBOLS, position)])
        assert_array_equal(result, expected)

    @parametrize("position", [-3, 6, [0, 0, 0, 0, -5]])
    def test_char_failure(self, position):
        with pytest.raises(ValueError, match=r"Position -?\d out of bounds"):
            FAString(SYMBOLS).char(position)

    @parametrize(
        "str2, expected",
        [
            ("A", [True, True, False, False, False]),
            ("AA", [True, False, False, False, False]),
            ("", [True] * 5),
            ("AAA", [False] * 5),
            ("AAPL", [True] + [False] * 4),
        ],
    )
    def test_contains(self, str2, expected):
        result = FAString(SYMBOLS).contains(str2)
        assert_array_equal(result, expected)

        result = FAString(NB_PARALLEL_SYMBOLS).contains(str2)
        assert_array_equal(result, expected * PARALLEL_MULTIPLIER)

    @pytest.mark.skip(
        reason="Test adapted from `test_contains` but expected results need to be adapted for `cat_symbol` data."
    )
    @parametrize(
        "str2, expected",
        [
            ("A", [True, True, False, False, False]),
            ("AA", [True, False, False, False, False]),
            ("", [True] * 5),
            ("AAA", [False] * 5),
            ("AAPL", [True] + [False] * 4),
        ],
    )
    def test_contains_cat(self, str2, expected):
        result = self.cat_symbol.str.contains(str2)
        assert_array_equal(result, expected)

        # No parallel test for CatString. It internally uses FAString on the
        # category label array, and that code path is already tested elsewhere.

    @parametrize(
        "str2, expected",
        [
            ("A", [True, True, False, False, False]),
            ("AA", [True, False, False, False, False]),
            ("", [True] * 5),
            ("AAA", [False] * 5),
            ("AAPL", [True] + [False] * 4),
        ],
    )
    def test_strstrb(self, str2, expected):
        result = FAString(SYMBOLS).strstrb(str2)
        assert_array_equal(result, expected)

        result = FAString(NB_PARALLEL_SYMBOLS).strstrb(str2)
        assert_array_equal(result, expected * PARALLEL_MULTIPLIER)

    @pytest.mark.skip(
        reason="Test adapted from `test_contains` but expected results need to be adapted for `cat_symbol` data."
    )
    @parametrize(
        "str2, expected",
        [
            ("A", [True, True, False, False, False]),
            ("AA", [True, False, False, False, False]),
            ("", [True] * 5),
            ("AAA", [False] * 5),
            ("AAPL", [True] + [False] * 4),
        ],
    )
    def test_strstrb_cat(self, str2, expected):
        result = self.cat_symbol.str.strstrb(str2)
        assert_array_equal(result, expected)

        # No parallel test for CatString. It internally uses FAString on the
        # category label array, and that code path is already tested elsewhere.

    @parametrize(
        "func, expected, filtered",
        [
            ("index", 0, rt.rt_enum.INVALID_POINTER_32),
            ("index_any_of", 0, rt.rt_enum.INVALID_POINTER_32),
            ("contains", True, False),
            ("startswith", True, False),
            ("endswith", True, False),
        ],
    )
    def test_empty_string_comparisons_cat(self, func, expected, filtered):
        result = getattr(self.cat_symbol.str, func)("")
        expected = rt.tile(expected, len(self.cat_symbol)).astype(result.dtype)
        expected[self.cat_symbol.isfiltered()] = filtered
        assert_array_equal(result, expected)

    @parametrize(
        "str2, expected",
        [
            ("bb", [False, False, True]),
            ("ba", [False, True, False]),
        ],
    )
    def test_endswith(self, str2, expected):
        # TODO: Expand test to include when the underlying array is a Categorical.
        #       Also test parallel case.
        result = FAString(["abab", "ababa", "abababb"]).endswith(str2)
        assert_array_equal(result, expected)

    @parametrize(
        "str2, expected",
        [
            ("A", [0, 0, -1, -1, -1]),
            ("AA", [0, -1, -1, -1, -1]),
            ("AAPL", [0, -1, -1, -1, -1]),
            ("", [0] * 5),
            ("AAA", [-1] * 5),
            ("B", [-1, -1, 1, -1, 1]),
        ],
    )
    def test_index(self, str2, expected):
        result = FAString(SYMBOLS).index(str2)
        assert_array_equal(result, expected)

        result = FAString(NB_PARALLEL_SYMBOLS).index(str2)
        assert_array_equal(result, expected * PARALLEL_MULTIPLIER)

        # test old alias
        result = FAString(SYMBOLS).strstr(str2)
        assert_array_equal(result, expected)

    def test_index_cat(self):
        result = self.cat_symbol.str.index("A")
        inv = rt.INVALID_DICT[np.dtype(result.dtype).num]
        expected = rt.FA([inv, 0, 0, -1, -1, -1], dtype=result.dtype).tile(3)
        assert_array_equal(result, expected)

    def test_index_any_of(self):
        result = FAString(SYMBOLS).index_any_of("PZG")
        expected = rt.FA([2, 2, -1, 0, -1])
        assert_array_equal(result, expected)

        result = FAString(NB_PARALLEL_SYMBOLS).index_any_of("PZG")
        expected = rt.FA([2, 2, -1, 0, -1] * PARALLEL_MULTIPLIER)
        assert_array_equal(result, expected)

        # test old alias
        result = FAString(SYMBOLS).strpbrk("PZG")
        expected = rt.FA([2, 2, -1, 0, -1])
        assert_array_equal(result, expected)

    def test_index_any_of_cat(self):
        result = self.cat_symbol.str.index_any_of("PZG")
        inv = rt.INVALID_DICT[np.dtype(result.dtype).num]
        expected = rt.FA([inv, 2, 2, -1, 0, -1], dtype=result.dtype).tile(3)
        assert_array_equal(result, expected)

    def test_lower(self):
        result = FAString(SYMBOLS).lower
        assert result.tolist() == [s.lower() for s in SYMBOLS]

    def test_lower_cat(self):
        result = self.cat_symbol.str.lower
        expected = Cat(self.cat_symbol.ikey, [s.lower() for s in SYMBOLS])
        assert_array_or_cat_equal(result, expected, relaxed_cat_check=True)

    regex_match_test_cases = parametrize(
        "regex, expected",
        [
            (".", [True] * 5),
            (r"\.", [False] * 5),
            ("A", [True, True, False, False, False]),
            ("[A|B]", [True, True, True, False, True]),
            ("B$", [False, False, True, False, False]),
        ],
    )

    @parametrize("apply_unique", [True, False])
    @regex_match_test_cases
    def test_regex_match(self, regex, expected, apply_unique):
        fa = FA(SYMBOLS * 2)
        assert_array_equal(fa.str.regex_match(regex, apply_unique=apply_unique), expected * 2)

    @regex_match_test_cases
    def test_regex_match_cat(self, regex, expected):
        cat = Cat(SYMBOLS * 2)  # introduce duplicity to test ikey properly
        assert_array_equal(cat.str.regex_match(regex), expected * 2)

    regex_replace_test_cases = parametrize(
        "regex, repl",
        [
            (".", "A"),
            ("\w+", "A"),
            ("\d+", "0"),
            ("[A|B|C]", "C"),
            ("B$", ""),
        ],
    )

    @pytest.mark.parametrize(
        "arr_type_factory", [pytest.param(rt.FastArray, id="FastArray"), pytest.param(rt.Categorical, id="Categorical")]
    )
    @parametrize("apply_unique", [True, False])
    @regex_replace_test_cases
    def test_regex_replace(self, regex, repl, apply_unique, arr_type_factory):
        fa = arr_type_factory(SYMBOLS * 2)
        result = fa.str.regex_replace(regex, repl, apply_unique=apply_unique)
        expected = arr_type_factory([re.sub(regex, repl, s) for s in SYMBOLS * 2])
        assert_array_or_cat_equal(result, expected)

    def test_removetrailing_empty(self) -> None:
        arr = rt.FA([], dtype="S11")  # empty array
        result = arr.str.removetrailing()
        assert_array_equal(arr, result)

    @pytest.mark.parametrize(
        "strings",
        [
            pytest.param(
                [
                    (x + (" " * (((len(x) * 17) + idx) % 3)) if idx % 3 != 0 else x)
                    for idx, x in enumerate(_top_25_US_cities_by_population_2019_state_names)
                ],
                id="modified US state names",
            )
        ],
    )
    @pytest.mark.parametrize(
        "arr_type_factory", [pytest.param(rt.FastArray, id="FastArray"), pytest.param(rt.Categorical, id="Categorical")]
    )
    @pytest.mark.parametrize("unicode", [pytest.param(False, id="ascii"), pytest.param(True, id="unicode")])
    @pytest.mark.parametrize("parallel", [pytest.param(False, id="serial"), pytest.param(True, id="parallel")])
    @pytest.mark.parametrize("chars", [pytest.param(" ", id="space"), "s"])
    def test_removetrailing(
        self, strings: list, arr_type_factory: callable, unicode: bool, parallel: bool, chars
    ) -> None:
        dtype_str = "<U" if unicode else "|S"
        if not unicode:
            strings = [x.encode() for x in strings]
        ndarray = np.array(strings, dtype=dtype_str)
        arr = arr_type_factory(ndarray)
        arr, tile_count = (arr, None) if not parallel else _make_parallelizable_array(arr)

        # Perform the operation on the string array.
        result = arr.str.removetrailing(remove=ord(chars))

        # Perform the equivalent operation in pure Python.
        chars = chars if unicode else chars.encode()
        expected_strings = [x.strip(chars) for x in strings]
        expected_ndarray = np.array(expected_strings, dtype=dtype_str)
        expected = arr_type_factory(expected_ndarray)
        if tile_count is not None:
            expected = _tile_array(expected, tile_count)

        # Verify the output matches the expected result.
        assert_array_or_cat_equal(
            result,
            expected,
            # TEMP: The function doesn't shrink the dtype of the category array in
            #       an output categorical. So the dtype of the output cat will be the
            #       same as that of the output cat, even if trailing character(s) have
            #       been removed and the strings could fit in a smaller dtype.
            exact_dtype_match=False,
        )

    #
    # TODO: reverse tests
    #

    #
    # TODO: reverse_inplace tests
    #

    def test_find(self):
        res = FAString(["this", "that", "test"])._find("t")
        expected = FastArray([[True, False, False, False], [True, False, False, True], [True, False, False, True]])
        assert_array_equal(res, expected)

    @pytest.mark.parametrize(
        "arr_type_factory", [pytest.param(rt.FastArray, id="FastArray"), pytest.param(rt.Categorical, id="Categorical")]
    )
    @pytest.mark.parametrize("unicode", [pytest.param(False, id="ascii"), pytest.param(True, id="unicode")])
    @pytest.mark.parametrize("parallel", [pytest.param(False, id="serial"), pytest.param(True, id="parallel")])
    @pytest.mark.parametrize(
        "old, new",
        [("A", "B"), ("A", "BB"), ("A", ""), ("OO", "O"), ("FB", "TWITTER"), ("AAPL", ""), ("XYZ", "LMNOP")],
        ids=[
            "one-for-one",
            "one-for-two",
            "one-for-none",
            "two-for-one-pair",
            "all-replaced",
            "all-replaced-with-empty",
            "no-op",
        ],
    )
    def test_replace(self, arr_type_factory, unicode, parallel, old, new):
        dtype_str = "<U" if unicode else "|S"
        ndarray = np.array(SYMBOLS, dtype=dtype_str)
        arr = arr_type_factory(ndarray)
        arr, tile_count = (arr, None) if not parallel else _make_parallelizable_array(arr)
        result = arr.str.replace(old, new)

        expected_strings = [x.replace(old, new) for x in SYMBOLS]
        expected_ndarray = np.array(expected_strings, dtype=dtype_str)
        expected = arr_type_factory(expected_ndarray)
        if tile_count is not None:
            expected = _tile_array(expected, tile_count)
        # breakpoint()
        assert_array_or_cat_equal(
            result,
            expected,
            exact_dtype_match=False,
            relaxed_cat_check=True,
        )

    def test_startswith(self):
        arrsize = 200
        symbol = Cat(1 + arange(arrsize) % len(SYMBOLS), SYMBOLS)
        assert_array_equal(symbol.expand_array.str.startswith("AAPL"), symbol.str.startswith("AAPL"))

    def test_startswith_cat(self):
        arrsize = 200
        symbol = Cat(1 + arange(arrsize) % len(SYMBOLS), SYMBOLS)
        assert_array_equal(symbol.expand_array.str.startswith("AAPL"), symbol.str.startswith("AAPL"))

    def test_startswith_cat_filtered(self):
        assert_array_equal(self.cat_symbol.expand_array.str.startswith("IBM"), self.cat_symbol.str.startswith("IBM"))

    def test_strlen_cat(self):
        result = self.cat_symbol.str.strlen
        inv = rt.INVALID_DICT[np.dtype(result.dtype).num]
        expected = rt.FA([inv, 4, 4, 2, 4, 3], dtype=result.dtype).tile(3)
        assert_array_equal(result, expected)

    def test_strlen_parallel(self):
        result = rt.FAString(NB_PARALLEL_SYMBOLS).strlen
        expected = rt.FA([4, 4, 2, 4, 3], dtype=result.dtype).tile(PARALLEL_MULTIPLIER)
        assert_array_equal(result, expected)

    def test_strpbrk_cat(self):
        result = self.cat_symbol.str.strpbrk("PZG")
        inv = rt.INVALID_DICT[np.dtype(result.dtype).num]
        expected = rt.FA([inv, 2, 2, -1, 0, -1], dtype=result.dtype).tile(3)
        assert_array_equal(result, expected)

    @parametrize(
        "str2, expected",
        [
            ("A", [0, 0, -1, -1, -1]),
            ("AA", [0, -1, -1, -1, -1]),
            ("AAPL", [0, -1, -1, -1, -1]),
            ("", [0] * 5),
            ("AAA", [-1] * 5),
            ("B", [-1, -1, 1, -1, 1]),
        ],
    )
    def test_strstr(self, str2, expected):
        result = FAString(SYMBOLS).strstr(str2)
        assert_array_equal(result, expected)

        result = FAString(NB_PARALLEL_SYMBOLS).strstr(str2)
        assert_array_equal(result, expected * PARALLEL_MULTIPLIER)

    def test_strstr_cat(self):
        result = self.cat_symbol.str.strstr("A")
        inv = rt.INVALID_DICT[np.dtype(result.dtype).num]
        expected = rt.FA([inv, 0, 0, -1, -1, -1], dtype=result.dtype).tile(3)
        assert_array_equal(result, expected)

    substr_test_cases = parametrize(
        "start_stop",
        [
            (0, 1),
            (0, 2),
            (1, 3),
            (1, -1),
            (-2, 3),
            (2,),
            (-1,),
            (-3, -1),
            (-1, 1),
        ],
    )

    @substr_test_cases
    def test_substr(self, start_stop):
        expected = rt.FastArray([s[slice(*start_stop)] for s in SYMBOLS])
        result = FAString(SYMBOLS).substr(*start_stop)
        assert_array_equal(expected, result)

    @substr_test_cases
    def test_substr_bytes(self, start_stop):
        expected = [s[slice(*start_stop)] for s in SYMBOLS]
        result = FAString(FastArray(SYMBOLS)).substr(*start_stop)
        assert_array_equal(FastArray(expected), result)

    @substr_test_cases
    def test_substr_cat(self, start_stop):
        result = self.cat_symbol.str.substr(*start_stop)
        expected = [s[slice(*start_stop)] for s in SYMBOLS]
        expected = Categorical(self.cat_symbol.ikey, expected, base_index=1)
        assert_array_or_cat_equal(expected, result, relaxed_cat_check=True)
        # check categories are unique
        assert len(set(result.category_array)) == len(result.category_array)

    @parametrize(
        "start, stop, expected",
        [
            (0, [1, 2, 3, 2, 3], ["A", "AM", "FB", "GO", "IBM"]),
            ([1, 1, 1, 1, 1], [1, 2, 3, 2, 3], ["", "M", "B", "O", "BM"]),
            ([1, 2, 3, 2, 3], None, ["A", "AM", "FB", "GO", "IBM"]),
            ([0, 1, 1, 0, 1], [3, 10, 2, -1, -2], ["AAP", "MZN", "B", "GOO", ""]),
            ([1, 1, 1, 1, 1], [1, 1, 1, 1, 1], ["", "", "", "", ""]),
        ],
    )
    def test_substr_array_bounds(self, start, stop, expected):
        result = FAString(SYMBOLS).substr(start, stop)
        assert_array_equal(rt.FastArray(expected), result)

    @substr_test_cases
    def test_substr_getitem(self, start_stop):
        expected = rt.FastArray([s[slice(*start_stop)] for s in SYMBOLS])
        result = FAString(SYMBOLS).substr[slice(*start_stop)]
        assert_array_equal(expected, result)

    def test_substr_getitem_single(self):
        expected = rt.FastArray([s[0] for s in SYMBOLS])
        result = FAString(SYMBOLS).substr[0]
        assert_array_equal(expected, result)

    def test_substr_getitem_array(self):
        indexer = [0, 1, 0, 1, 0]
        expected = rt.FastArray([s[i] for s, i in zip(SYMBOLS, indexer)])
        result = FAString(SYMBOLS).substr[indexer]
        assert_array_equal(expected, result)

    def test_substr_char_stop(self):
        s = FastArray(["ABC", "A_B", "AB_C", "AB_C_DD"])
        res = s.str.substr_char_stop("_")
        expected = FastArray([b"ABC", b"A", b"AB", b"AB"], dtype="|S3")
        assert_array_equal(expected, res)

        res = s.str.substr_char_stop("_", inclusive=True)
        expected = FastArray([b"ABC", b"A_", b"AB_", b"AB_"], dtype="|S3")
        assert_array_equal(expected, res)

    def test_upper(self):
        result = FAString(SYMBOLS).upper
        assert result.tolist() == [s.upper() for s in SYMBOLS]

    def test_upper_cat(self):
        result = self.cat_symbol.str.upper
        expected = Cat(self.cat_symbol.ikey, [s.upper() for s in SYMBOLS])
        assert_array_or_cat_equal(result, expected, relaxed_cat_check=True)

    #
    # TODO: upper_inplace tests
    #


class TestExtract:
    duplicity = 2
    osi = rt.FastArray(
        [
            "SPX UO 12/15/23 C5700",
            "SPX UO 07/16/21 P3480",
            "SPXW UO 07/16/21 P3190",
            "SPXW UO 06/30/21 C4100",
            "SPXW UO 09/17/21 C3650",
        ]
        * duplicity
    )

    expirations = [b"12/15/23", b"07/16/21", b"07/16/21", b"06/30/21", b"09/17/21"] * duplicity
    roots = [b"SPX", b"SPX", b"SPXW", b"SPXW", b"SPXW"] * duplicity
    strikes = [b"5700", b"3480", b"3190", b"4100", b"3650"] * duplicity

    dataset_out_test_cases = parametrize(
        "pattern, expected",
        [
            ("(\w+).* (\d{2}/\d{2}/\d{2})", dict(group_0=roots, group_1=expirations)),
            ("(?P<root>\w+).*(\d{2}/\d{2}/\d{2})", dict(root=roots, group_1=expirations)),
            ("(?P<root>\w+).*(?P<expiration>\d{2}/\d{2}/\d{2})", dict(root=roots, expiration=expirations)),
            (" [C|P](?P<strike>\d+)$", dict(strike=strikes)),
            (
                "(?P<root>\w+W).*(?P<expiration>\d{2}/\d{2}/\d{2})",
                dict(
                    root=[root if b"W" in s else "" for s, root in zip(osi, roots)],
                    expiration=[exp if b"W" in s else "" for s, exp in zip(osi, expirations)],
                ),
            ),
        ],
        ids=["non-names", "some-names", "all-names", "single-named", "some-unmatched"],
    )

    @parametrize("apply_unique", [True, False])
    @dataset_out_test_cases
    def test_extract_dataset(self, pattern, expected, apply_unique):
        result = self.osi.str.extract(pattern, expand=True, apply_unique=apply_unique)
        assert result.keys() == list(expected)
        [
            assert_array_or_cat_equal(
                FastArray(expected[key]),
                result[key],
            )
            for key in result
        ]

    @parametrize("apply_unique", [True, False])
    def test_extract_dataset_detected_names(self, apply_unique):
        pattern = " [C|P](?P<strike>\d+)$"
        expected = dict(strike=self.strikes)
        result = self.osi.str.extract(pattern, apply_unique=apply_unique)
        assert result.keys() == list(expected)
        [
            assert_array_or_cat_equal(
                FastArray(expected[key]),
                result[key],
            )
            for key in result
        ]

    array_out_test_cases = parametrize(
        "pattern, expected",
        [
            (" [C|P](\d+)", strikes),
            ("\w{2}", [s[:2] for s in roots]),
        ],
        ids=["group", "no-group"],
    )

    @array_out_test_cases
    def test_extract_array(self, pattern, expected):
        result = self.osi.str.extract(pattern)
        expected = rt.FastArray(expected)
        assert_array_or_cat_equal(expected, result)

    @parametrize("kwargs, key", [(dict(expand=True), "group_0"), (dict(names=["extract"]), "extract")])
    @array_out_test_cases
    def test_single_group_datasets(self, pattern, expected, kwargs, key):
        result = self.osi.str.extract(pattern, **kwargs)
        assert isinstance(result, Dataset)
        assert result.keys() == [key]
        assert_array_equal(expected, result[key])

    @dataset_out_test_cases
    def test_categorical_extract_dataset(self, pattern, expected):
        result = rt.Cat(self.osi).str.extract(
            pattern,
            expand=True,
        )
        [
            assert_array_or_cat_equal(
                Categorical(expected[key]), result[key], relaxed_cat_check=True, check_cat_names=False
            )
            for key in result
        ]

    @array_out_test_cases
    def test_categorical_extract_array(self, pattern, expected):
        result = rt.Cat(self.osi).str.extract(pattern)
        assert_array_or_cat_equal(Categorical(expected), result, relaxed_cat_check=True, check_cat_names=False)
